"""Shared test fixtures and configuration."""

from __future__ import annotations

import json
import tempfile
from pathlib import Path
from typing import Any, Callable, Generator
from unittest.mock import Mock

import pytest


@pytest.fixture
def temp_json_file() -> Generator[Path, Any, None]:
    """Create a temporary JSON file for testing."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".json", delete=False) as f:
        temp_path = Path(f.name)
    yield temp_path
    if temp_path.exists():
        temp_path.unlink()


@pytest.fixture
def temp_config_file() -> Generator[Path, Any, None]:
    """Create a temporary config file for testing."""
    with tempfile.NamedTemporaryFile(mode="w", suffix=".config", delete=False) as f:
        temp_path = Path(f.name)
    yield temp_path
    if temp_path.exists():
        temp_path.unlink()


@pytest.fixture
def sample_credentials() -> dict[str, str]:
    """Sample valid Snowflake credentials for testing."""
    return {
        "account": "test_account",
        "user": "test_user",
        "password": "test_password",
        "database": "test_db",
        "schema": "test_schema",
        "warehouse": "test_warehouse",
        "role": "test_role",
    }


@pytest.fixture
def sample_legacy_credentials() -> dict[str, str]:
    """Sample legacy format Snowflake credentials for testing."""
    return {
        "accountname": "test_account",
        "username": "test_user",
        "dbname": "test_db",
        "schemaname": "test_schema",
        "warehousename": "test_warehouse",
        "rolename": "test_role",
    }


@pytest.fixture
def mock_boto3_session() -> tuple[Mock, Mock]:
    """Mock boto3 session for AWS tests."""
    mock_session = Mock()
    mock_client = Mock()
    mock_session.client.return_value = mock_client
    return mock_session, mock_client


@pytest.fixture
def write_json_file(temp_json_file: Path) -> Callable[..., Any]:
    """Helper to write JSON data to a temporary file."""

    def _write_json(data: dict[str, Any]) -> Path:
        temp_json_file.write_text(json.dumps(data))
        return temp_json_file

    return _write_json


@pytest.fixture
def write_config_file(temp_config_file: Path) -> Callable[..., Any]:
    """Helper to write config data to a temporary file."""

    def _write_config(section_data: dict[str, str], section_name: str = "connections") -> Path:
        lines = [f"[{section_name}]"]
        for key, value in section_data.items():
            lines.append(f"{key} = {value}")
        temp_config_file.write_text("\n".join(lines))
        return temp_config_file

    return _write_config



"""This module provides functions to build Snowflake connections using various methods.

Some of the methods:
- JSON files
- snowflake config files
- environment variables
- AWS Secrets Manager.
"""

from __future__ import annotations

import configparser
import json
import logging
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
from snowflake.connector.connection import DEFAULT_CONFIGURATION
from sqlalchemy import create_engine


if TYPE_CHECKING:
    import boto3
    from sqlalchemy.engine import Engine


logger = logging.getLogger(__name__)

AVAILABLE_PARAMS = {
    *DEFAULT_CONFIGURATION.keys(),
    "private_key_encrypted",
    "private_key_passphrase",
}

# Key aliases for Snowflake connection parameters
KEY_ALIASES = {
    # SnowSQL specific
    "accountname": "account",
    "username": "user",
    "dbname": "database",
    "schemaname": "schema",
    "warehousename": "warehouse",
    "rolename": "role",
    # other compatible mapping
    "rsa_private_key": "private_key_encrypted",
    "pk_passphrase": "private_key_passphrase",
}

CREDS_ENV_VARS_PREFIX = "SNOWFLAKE_"
SNOWFLAKE_SETTINGS_JSON_PATH = Path(
    os.environ.get("SNOWFLAKE_SETTINGS_JSON_PATH", "/etc/config/snowflake_creds.json"),
)
SNOWFLAKE_CONFIG_FILE_PATH = Path(
    os.environ.get("SNOWFLAKE_CONFIG_FILE_PATH", Path.home() / ".snowsql" / "config"),
)
RESERVED_ENV_KEYS = {
    "SNOWFLAKE_SETTINGS_JSON_PATH",
    "SNOWFLAKE_CONFIG_FILE_PATH",
}


class ConnectionError(Exception):
    """Base exception for connection-related errors."""


def load_from_json_file(file: Path = SNOWFLAKE_SETTINGS_JSON_PATH) -> dict[str, Any]:
    """Load snowflake json config file.

    Returns:
        dict[str, Any]: snowflake config
    """
    logger.debug(f"Trying read snowflake credentials from json config file {file}")
    try:
        creds_dict = json.loads(file.read_text())
    except FileNotFoundError:
        logger.debug("File not found.")
        return {}
    creds = _sanitize_snowflake_credentials(creds_dict)
    if creds:
        logger.info(f"Loaded snowflake credentials from json config file: {file}")
    else:
        logger.debug("No credentials found in json config file")
    return creds


def load_from_snowflake_config_file(
    file: Path = SNOWFLAKE_CONFIG_FILE_PATH,
    section: str = "connections",
) -> dict[str, Any]:
    """Load snowflake credentials from a config file.

    Returns:
        dict[str, Any]: snowflake config
    """
    logger.debug(f"Trying read snowflake credentials from config file {file}")
    config = configparser.ConfigParser()
    if not file.exists():
        logger.debug("File not found.")
        return {}
    config.read(file)
    try:
        creds = dict(config[section])
        creds = _sanitize_snowflake_credentials(creds)
        if creds:
            logger.info(f"Loaded snowflake credentials from snowflake config file: {file}")
        else:
            logger.debug("No credentials found in snowflake config file")
        return creds
    except KeyError:
        return {}


def load_from_env_vars() -> dict[str, Any]:
    """Load snowflake credentials from environment variables.

    This function looks for environment variables that start with the prefix
    `SNOWFLAKE_` and returns a dictionary of the credentials.

    Returns:
        dict[str, Any]: snowflake credentials
    """
    logger.debug(
        f"Trying read snowflake credentials from environment variables with prefix {CREDS_ENV_VARS_PREFIX}",
    )
    env_creds = {
        cred_key: cred_value
        for cred_key, cred_value in os.environ.items()
        if cred_key.startswith(CREDS_ENV_VARS_PREFIX)
    }
    creds = _sanitize_snowflake_credentials(env_creds)
    if creds:
        logger.info(
            f"Loaded snowflake credentials from environment variables with prefix {CREDS_ENV_VARS_PREFIX}"
        )
    else:
        logger.debug("No credentials found in environment variables")
    return creds


def load_from_aws_secret(secret_name: str, session: boto3.Session) -> dict[str, Any]:
    """Load Snowflake credentials from AWS Secrets Manager.

    Args:
        secret_name (str): The name of the secret in AWS Secrets Manager.
        session (boto3.Session): A boto3 session object.

    Returns:
        dict[str, Any]: Snowflake credentials
    """
    logger.debug(f"Trying to load snowflake credentials from AWS Secrets Manager: {secret_name}")
    client = session.client(service_name="secretsmanager")
    get_secret_value_response = client.get_secret_value(SecretId=secret_name)

    secret_str = get_secret_value_response.get("SecretString")
    if not secret_str:
        logger.debug(f"No secret string found for secret: {secret_name}")
        return {}
    try:
        data = json.loads(secret_str)
    except json.JSONDecodeError:
        logger.error(f"Failed to decode JSON from secret: {secret_name}")
        return {}
    creds = _sanitize_snowflake_credentials(data)
    if creds:
        logger.info(f"Loaded snowflake credentials from AWS Secrets Manager: {secret_name}")
    else:
        logger.debug("No credentials found in AWS Secrets Manager")
    return creds


# utilities
def _sanitize_snowflake_credentials(creds: dict[str, Any]) -> dict[str, Any]:
    """Sanitize and normalize snowflake credentials.

    This function filters out empty values and maps legacy keys to their current equivalents.
    It also handles encrypted private keys with passphrases.

    Args:
        creds (Dict[str, Any]): raw credentials

    Returns:
        Dict[str, Any]: normalized credentials
    """
    norm_creds = {}
    for cred_key, cred_val in creds.items():
        if cred_val is None or (isinstance(cred_val, str) and not cred_val.strip()):
            continue
        norm_key = cred_key.lower()
        if norm_key.startswith(CREDS_ENV_VARS_PREFIX.lower()):
            norm_key = norm_key[len(CREDS_ENV_VARS_PREFIX) :]
        norm_key = norm_key.strip()
        mapped_norm_key = KEY_ALIASES.get(norm_key, norm_key)
        if mapped_norm_key in AVAILABLE_PARAMS:
            norm_creds[mapped_norm_key] = cred_val

    # deal with encrypted private key strings
    _private_key_encrypted = norm_creds.pop("private_key_encrypted", None)
    _private_key_passphrase = norm_creds.pop("private_key_passphrase", None)
    if _private_key_encrypted and not _private_key_passphrase:
        raise ValueError("private_key_encrypted is set but private_key_passphrase is not provided")
    if _private_key_encrypted and _private_key_passphrase:
        norm_creds["private_key"] = _load_private_with_passphrase(
            _private_key_encrypted,
            _private_key_passphrase,
        )
    return norm_creds


def _load_private_with_passphrase(private_key_pem: str, passphrase: str) -> bytes:
    private_key = serialization.load_pem_private_key(
        private_key_pem.encode(),
        password=passphrase.encode(),
        backend=default_backend(),
    )

    # Convert to DER format (required by snowflake.connector)
    private_key_der = private_key.private_bytes(
        encoding=serialization.Encoding.DER,
        format=serialization.PrivateFormat.PKCS8,
        encryption_algorithm=serialization.NoEncryption(),
    )
    return private_key_der


def create_snowflake_sa_engine(creds: dict[str, Any]) -> Engine:
    """Create a SQLAlchemy engine for Snowflake using the provided credentials.

    Args:
        creds (Dict[str, Any]): Snowflake credentials.

    Returns:
        Engine: SQLAlchemy engine instance.

    Raises:
        ConnectionError: If the engine creation fails.
    """
    creds = _sanitize_snowflake_credentials(creds)
    url_param_keys = {"user", "password", "account", "database", "schema"}
    connect_args = {k: creds[k] for k in creds if k not in url_param_keys and creds[k] is not None}

    try:
        # Build connection URL
        account = creds["account"]
        user = creds["user"]
        password = creds.get("password", "")
        database = creds.get("database", "")
        schema = creds.get("schema", "")

        # Handle password in URL
        user_pass = f"{user}:{password}" if password else user

        # Build URL
        url_parts = [f"snowflake://{user_pass}@{account}"]
        if database:
            url_parts.append(f"/{database}")

        # Add query parameters
        query_params = []
        if schema:
            query_params.append(f"schema={schema}")

        if query_params:
            url_parts.append("?" + "&".join(query_params))

        connection_url = "".join(url_parts)

        logger.debug("Creating SQLAlchemy engine")
        return create_engine(connection_url, connect_args=connect_args)
    except Exception as e:
        raise ConnectionError(f"Failed to create SQLAlchemy engine: {e}") from e



"""Tests for the connection_builder module."""

from __future__ import annotations

import json
import os
from unittest.mock import Mock, patch

import pytest
from snowconn.connection_builder import (
    ConnectionError,
    _load_private_with_passphrase,
    _sanitize_snowflake_credentials,
    create_snowflake_sa_engine,
    load_from_aws_secret,
    load_from_env_vars,
    load_from_json_file,
    load_from_snowflake_config_file,
)
from sqlalchemy.engine import Engine


class TestLoadFromJsonFile:
    """Test cases for load_from_json_file function."""

    def test_load_from_json_file_success(self, write_json_file, sample_credentials):
        """Test successful loading of JSON credentials."""
        json_file = write_json_file(sample_credentials)
        result = load_from_json_file(json_file)
        assert result == sample_credentials

    def test_load_from_json_file_not_found(self, tmp_path):
        """Test handling of missing JSON file."""
        non_existent_file = tmp_path / "non_existent.json"
        result = load_from_json_file(non_existent_file)
        assert result == {}

    def test_load_from_json_file_invalid_json(self, tmp_path):
        """Test handling of invalid JSON content."""
        json_file = tmp_path / "invalid.json"
        json_file.write_text("invalid json content")
        with pytest.raises(json.JSONDecodeError):
            load_from_json_file(json_file)

    def test_load_from_json_file_empty_credentials(self, write_json_file):
        """Test handling of empty or null credentials."""
        test_creds = {
            "account": "",
            "user": None,
            "password": "   ",  # whitespace only
            "database": "test_db",
        }
        json_file = write_json_file(test_creds)
        result = load_from_json_file(json_file)
        # Only database should remain after sanitization
        assert result == {"database": "test_db"}

    def test_load_from_json_file_legacy_keys(self, write_json_file, sample_legacy_credentials):
        """Test mapping of legacy keys to current format."""
        json_file = write_json_file(sample_legacy_credentials)
        result = load_from_json_file(json_file)
        expected = {
            "account": "test_account",
            "user": "test_user",
            "database": "test_db",
            "schema": "test_schema",
            "warehouse": "test_warehouse",
            "role": "test_role",
        }
        assert result == expected


class TestLoadFromSnowflakeConfigFile:
    """Test cases for load_from_snowflake_config_file function."""

    def test_load_from_config_file_success(self, write_config_file, sample_credentials):
        """Test successful loading of config file credentials."""
        config_file = write_config_file(sample_credentials)
        result = load_from_snowflake_config_file(config_file)
        assert result == sample_credentials

    def test_load_from_config_file_not_found(self, tmp_path):
        """Test handling of missing config file."""
        non_existent_file = tmp_path / "non_existent_config"
        result = load_from_snowflake_config_file(non_existent_file)
        assert result == {}

    def test_load_from_config_file_missing_section(self, write_config_file):
        """Test handling of missing section in config file."""
        config_file = write_config_file({"account": "test_account"}, section_name="other_section")
        result = load_from_snowflake_config_file(config_file, section="connections")
        assert result == {}

    def test_load_from_config_file_custom_section(self, write_config_file):
        """Test loading from custom section name."""
        test_creds = {"account": "test_account", "user": "test_user"}
        config_file = write_config_file(test_creds, section_name="custom_connections")
        result = load_from_snowflake_config_file(config_file, section="custom_connections")
        assert result == test_creds


class TestLoadFromEnvVars:
    """Test cases for load_from_env_vars function."""

    def test_load_from_env_vars_success(self):
        """Test successful loading of environment variables."""
        env_vars = {
            "SNOWFLAKE_ACCOUNT": "test_account",
            "SNOWFLAKE_USER": "test_user",
            "SNOWFLAKE_PASSWORD": "test_password",
            "SNOWFLAKE_DATABASE": "test_db",
            "OTHER_VAR": "should_be_ignored",
        }

        with patch.dict(os.environ, env_vars, clear=True):
            result = load_from_env_vars()

        expected = {
            "account": "test_account",
            "user": "test_user",
            "password": "test_password",
            "database": "test_db",
        }
        assert result == expected

    def test_load_from_env_vars_empty(self):
        """Test handling when no Snowflake environment variables are set."""
        env_vars = {
            "OTHER_VAR": "should_be_ignored",
            "ANOTHER_VAR": "also_ignored",
        }

        with patch.dict(os.environ, env_vars, clear=True):
            result = load_from_env_vars()

        assert result == {}

    def test_load_from_env_vars_case_insensitive(self):
        """Test that environment variable processing is case insensitive."""
        env_vars = {
            "SNOWFLAKE_ACCOUNT": "test_account",
            "snowflake_user": "test_user",  # lowercase should still work
        }

        with patch.dict(os.environ, env_vars, clear=True):
            result = load_from_env_vars()

        # Note: actual env vars are case sensitive, but the processing should handle it
        assert "account" in result
        assert result["account"] == "test_account"


class TestLoadFromAwsSecret:
    """Test cases for load_from_aws_secret function."""

    def test_load_from_aws_secret_success(self, mock_boto3_session, sample_credentials):
        """Test successful loading from AWS Secrets Manager."""
        mock_session, mock_client = mock_boto3_session
        mock_client.get_secret_value.return_value = {"SecretString": json.dumps(sample_credentials)}

        result = load_from_aws_secret("test_secret", mock_session)

        assert result == sample_credentials
        mock_session.client.assert_called_once_with(service_name="secretsmanager")
        mock_client.get_secret_value.assert_called_once_with(SecretId="test_secret")

    def test_load_from_aws_secret_no_secret_string(self, mock_boto3_session):
        """Test handling when secret has no SecretString."""
        mock_session, mock_client = mock_boto3_session
        mock_client.get_secret_value.return_value = {}

        result = load_from_aws_secret("test_secret", mock_session)

        assert result == {}

    def test_load_from_aws_secret_invalid_json(self, mock_boto3_session):
        """Test handling of invalid JSON in secret."""
        mock_session, mock_client = mock_boto3_session
        mock_client.get_secret_value.return_value = {"SecretString": "invalid json"}

        result = load_from_aws_secret("test_secret", mock_session)

        assert result == {}

    def test_load_from_aws_secret_client_error(self, mock_boto3_session):
        """Test handling of AWS client errors."""
        mock_session, mock_client = mock_boto3_session
        mock_client.get_secret_value.side_effect = Exception("AWS Error")

        with pytest.raises(Exception, match="AWS Error"):
            load_from_aws_secret("test_secret", mock_session)


class TestSanitizeSnowflakeCredentials:
    """Test cases for _sanitize_snowflake_credentials function."""

    def test_sanitize_basic_credentials(self):
        """Test basic credential sanitization."""
        raw_creds = {
            "SNOWFLAKE_ACCOUNT": "test_account",
            "SNOWFLAKE_USER": "test_user",
            "password": "test_password",
            "database": "",  # empty string should be filtered out
            "schema": None,  # None should be filtered out
            "warehouse": "   ",  # whitespace only should be filtered out
            "role": "test_role",
        }

        result = _sanitize_snowflake_credentials(raw_creds)

        expected = {
            "account": "test_account",
            "user": "test_user",
            "password": "test_password",
            "role": "test_role",
        }
        assert result == expected

    def test_sanitize_legacy_keys(self):
        """Test mapping of legacy keys."""
        raw_creds = {
            "accountname": "test_account",
            "username": "test_user",
            "dbname": "test_db",
            "schemaname": "test_schema",
            "warehousename": "test_warehouse",
            "rolename": "test_role",
        }

        result = _sanitize_snowflake_credentials(raw_creds)

        expected = {
            "account": "test_account",
            "user": "test_user",
            "database": "test_db",
            "schema": "test_schema",
            "warehouse": "test_warehouse",
            "role": "test_role",
        }
        assert result == expected

    def test_sanitize_env_var_prefix_removal(self):
        """Test removal of SNOWFLAKE_ prefix from environment variables."""
        raw_creds = {
            "SNOWFLAKE_ACCOUNT": "test_account",
            "SNOWFLAKE_USER": "test_user",
            "snowflake_password": "test_password",  # lowercase prefix
        }

        result = _sanitize_snowflake_credentials(raw_creds)

        expected = {
            "account": "test_account",
            "user": "test_user",
            "password": "test_password",
        }
        assert result == expected

    def test_sanitize_invalid_params_filtered(self):
        """Test that invalid parameters are filtered out."""
        raw_creds = {
            "account": "test_account",
            "invalid_param": "should_be_filtered",
            "another_invalid": "also_filtered",
            "user": "test_user",
        }

        result = _sanitize_snowflake_credentials(raw_creds)

        expected = {
            "account": "test_account",
            "user": "test_user",
        }
        assert result == expected

    @patch("snowconn.connection_builder._load_private_with_passphrase")
    def test_sanitize_encrypted_private_key_success(self, mock_load_private):
        """Test handling of encrypted private key with passphrase."""
        mock_private_key = b"decoded_private_key"
        mock_load_private.return_value = mock_private_key

        raw_creds = {
            "account": "test_account",
            "user": "test_user",
            "private_key_encrypted": "encrypted_key_content",
            "private_key_passphrase": "passphrase",
        }

        result = _sanitize_snowflake_credentials(raw_creds)

        expected = {
            "account": "test_account",
            "user": "test_user",
            "private_key": mock_private_key,
        }
        assert result == expected
        mock_load_private.assert_called_once_with("encrypted_key_content", "passphrase")

    def test_sanitize_encrypted_private_key_missing_passphrase(self):
        """Test error when encrypted private key is provided without passphrase."""
        raw_creds = {
            "account": "test_account",
            "private_key_encrypted": "encrypted_key_content",
            # missing private_key_passphrase
        }

        with pytest.raises(
            ValueError,
            match="private_key_encrypted is set but private_key_passphrase is not provided",
        ):
            _sanitize_snowflake_credentials(raw_creds)


class TestLoadPrivateWithPassphrase:
    """Test cases for _load_private_with_passphrase function."""

    @patch("snowconn.connection_builder.serialization.load_pem_private_key")
    def test_load_private_key_success(self, mock_load_pem):
        """Test successful loading and conversion of private key."""
        # Mock the private key object
        mock_private_key = Mock()
        mock_private_key_der = b"der_encoded_key"
        mock_private_key.private_bytes.return_value = mock_private_key_der
        mock_load_pem.return_value = mock_private_key

        private_key_pem = "-----BEGIN PRIVATE KEY-----\ntest_key_content\n-----END PRIVATE KEY-----"
        passphrase = "test_passphrase"

        result = _load_private_with_passphrase(private_key_pem, passphrase)

        assert result == mock_private_key_der
        mock_load_pem.assert_called_once()
        # Verify the arguments passed to load_pem_private_key
        args, kwargs = mock_load_pem.call_args
        assert args[0] == private_key_pem.encode()
        assert kwargs["password"] == passphrase.encode()

    @patch("snowconn.connection_builder.serialization.load_pem_private_key")
    def test_load_private_key_invalid_key(self, mock_load_pem):
        """Test handling of invalid private key."""
        mock_load_pem.side_effect = ValueError("Invalid private key")

        with pytest.raises(ValueError, match="Invalid private key"):
            _load_private_with_passphrase("invalid_key", "passphrase")


class TestCreateSnowflakeSaEngine:
    """Test cases for create_snowflake_sa_engine function."""

    @patch("snowconn.connection_builder.create_engine")
    def test_create_engine_success(self, mock_create_engine):
        """Test successful SQLAlchemy engine creation."""
        mock_engine = Mock(spec=Engine)
        mock_create_engine.return_value = mock_engine

        creds = {
            "account": "test_account",
            "user": "test_user",
            "password": "test_password",
            "database": "test_db",
            "schema": "test_schema",
            "warehouse": "test_warehouse",
            "role": "test_role",
        }

        result = create_snowflake_sa_engine(creds)

        assert result == mock_engine
        mock_create_engine.assert_called_once()

        # Verify the connection URL format
        args, kwargs = mock_create_engine.call_args
        connection_url = args[0]
        assert "snowflake://test_user:test_password@test_account/test_db" in connection_url
        assert "schema=test_schema" in connection_url

        # Verify connect_args contains non-URL parameters
        connect_args = kwargs["connect_args"]
        assert connect_args["warehouse"] == "test_warehouse"
        assert connect_args["role"] == "test_role"

    @patch("snowconn.connection_builder.create_engine")
    def test_create_engine_without_password(self, mock_create_engine):
        """Test engine creation without password."""
        mock_engine = Mock(spec=Engine)
        mock_create_engine.return_value = mock_engine

        creds = {
            "account": "test_account",
            "user": "test_user",
            "database": "test_db",
        }

        result = create_snowflake_sa_engine(creds)

        assert result == mock_engine
        args, kwargs = mock_create_engine.call_args
        connection_url = args[0]
        # Should not have password in URL
        assert "snowflake://test_user@test_account/test_db" in connection_url

    @patch("snowconn.connection_builder.create_engine")
    def test_create_engine_minimal_creds(self, mock_create_engine):
        """Test engine creation with minimal credentials."""
        mock_engine = Mock(spec=Engine)
        mock_create_engine.return_value = mock_engine

        creds = {
            "account": "test_account",
            "user": "test_user",
        }

        result = create_snowflake_sa_engine(creds)

        assert result == mock_engine
        args, kwargs = mock_create_engine.call_args
        connection_url = args[0]
        assert "snowflake://test_user@test_account" in connection_url

    @patch("snowconn.connection_builder.create_engine")
    def test_create_engine_failure(self, mock_create_engine):
        """Test handling of engine creation failure."""
        mock_create_engine.side_effect = Exception("Engine creation failed")

        creds = {
            "account": "test_account",
            "user": "test_user",
        }

        with pytest.raises(
            ConnectionError, match="Failed to create SQLAlchemy engine: Engine creation failed"
        ):
            create_snowflake_sa_engine(creds)


class TestIntegration:
    """Integration tests combining multiple functions."""

    def test_full_credential_flow_json(self, write_json_file, sample_credentials):
        """Test complete flow from JSON file to engine creation."""
        json_file = write_json_file(sample_credentials)
        creds = load_from_json_file(json_file)

        with patch("snowconn.connection_builder.create_engine") as mock_create_engine:
            mock_engine = Mock(spec=Engine)
            mock_create_engine.return_value = mock_engine

            engine = create_snowflake_sa_engine(creds)

            assert engine == mock_engine
            mock_create_engine.assert_called_once()

    def test_credential_precedence_and_merging(self, write_json_file):
        """Test that different credential sources can be combined."""
        # Test JSON file
        json_creds = {"account": "json_account", "user": "json_user"}
        json_file = write_json_file(json_creds)

        # Test environment variables
        env_vars = {
            "SNOWFLAKE_PASSWORD": "env_password",
            "SNOWFLAKE_DATABASE": "env_database",
        }

        json_result = load_from_json_file(json_file)

        with patch.dict(os.environ, env_vars, clear=True):
            env_result = load_from_env_vars()

        # Verify that each source provides its expected values
        assert json_result == {"account": "json_account", "user": "json_user"}
        assert env_result == {"password": "env_password", "database": "env_database"}

        # In a real application, these would be merged by the connection manager
        merged_creds = {**json_result, **env_result}
        expected_merged = {
            "account": "json_account",
            "user": "json_user",
            "password": "env_password",
            "database": "env_database",
        }
        assert merged_creds == expected_merged



"""Provides the SnowConn class for connecting to Snowflake using different credentials sources.

Supports executing queries, reading/writing pandas DataFrames, and managing connections.

Usage:
    conn = SnowConn.connect(methods=['local'])
    results = conn.execute_simple("SELECT * FROM my_table;")
    df = conn.read_df("SELECT * FROM my_table;")
    conn.write_df(df, table="MY_TABLE", schema="PUBLIC")
    conn.close()
"""

from __future__ import annotations

import json
import logging
import warnings
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, Iterable, Literal

import snowflake.connector

from snowconn.connection_builder import (
    SNOWFLAKE_CONFIG_FILE_PATH,
    create_snowflake_sa_engine,
    load_from_aws_secret,
    load_from_snowflake_config_file,
)


if TYPE_CHECKING:
    import pandas as pd
    from snowflake.connector import SnowflakeConnection
    from sqlalchemy.engine.base import Connection, Engine


class InvalidMethodException(Exception):  # noqa: N818
    """Exception raised for invalid connection methods."""


class SnowConn:
    """SnowConn is a class that provides methods to connect to Snowflake using various methods."""

    _alchemy_engine = None
    _connection = None
    _raw_connection = None

    def __init__(self) -> None:
        self._alchemy_engine = None
        self._connection = None
        self._raw_connection = None

    def __enter__(self) -> SnowConn:
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_value: BaseException | None,
        traceback: object | None,
    ) -> None:
        self.close()

    @classmethod
    def connect(
        cls,
        methods: Iterable[Literal["local", "secretsmanager", "credentials"]] = ("local",),
        *args: Any,
        **kwargs: Any,
    ) -> SnowConn:
        """Generic connect method.

        Will iterate through a list of connection methods until one succeeds
        in creating a connection
        from snowconn import SnowConn
        conn = SnowConn.autoconnect(method=['secretsmanager'], credsman_name='acme')
        """
        available_methods: dict[str, Callable[..., SnowConn]] = {
            "secretsmanager": cls.connect_secretsmanager,
            "local": cls.connect_local,
            "credentials": cls.connect_credentials,
        }
        for method in methods:
            if method in available_methods:
                try:
                    return available_methods[method](*args, **kwargs)
                except Exception as e:
                    logging.error(e)
        raise InvalidMethodException(
            f'methods {methods} are not a valid connection methods. Valid methods are "secretsmanager, local, credentials"'
        )

    @classmethod
    def connect_local(  # noqa: PLR0913
        cls,
        db: str = "public",
        schema: str = "public",
        autocommit: bool = True,
        role: str | None = None,
        warehouse: str | None = None,
        local_creds_path: str | None = None,
        **kwargs: Any,
    ) -> SnowConn:
        """Creates an engine and connection using your local snowsql credentials."""
        connect_args = kwargs.get("connect_args", {})
        conn = cls()
        creds = conn._get_local_creds(local_creds_path)
        conn._create_engine(creds, db, schema, autocommit, role, warehouse, connect_args)
        return conn

    def _get_local_creds(self, local_creds_path: str | None = None) -> dict[str, Any]:
        snowsql_config = Path(local_creds_path) if local_creds_path else SNOWFLAKE_CONFIG_FILE_PATH
        return load_from_snowflake_config_file(
            file=snowsql_config,
            section="connections",
        )

    @classmethod
    def connect_secretsmanager(  # noqa: PLR0913
        cls,
        credsman_name: str,
        db: str = "public",
        schema: str = "public",
        autocommit: bool = True,
        role: str | None = None,
        aws_region_name: str = "eu-west-1",
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
        warehouse: str | None = None,
        region_name: str | None = None,
        **kwargs: Any,
    ) -> SnowConn:
        """Creates an engine and connection using credentials from AWS secrets manager."""
        try:
            import boto3  # noqa
        except ImportError as e:
            logging.warning("boto3 not installed, cannot execute connect_secretsmanager")
            raise e
        if region_name:
            warnings.warn(
                """
                The argument 'region_name' will be deprecated in future versions.
                Please use 'region_name' instead
                """,
                DeprecationWarning,
                stacklevel=2,
            )
            aws_region_name = region_name

        connect_args = kwargs.get("connect_args", {})

        conn = SnowConn()
        creds = conn._get_secretsmanager_creds(
            credsman_name, aws_region_name, aws_access_key_id, aws_secret_access_key
        )
        conn._create_engine(creds, db, schema, autocommit, role, warehouse, connect_args)
        return conn

    @classmethod
    def credsman_connect(cls, *args: Any, **kwargs: Any) -> SnowConn:
        """Legacy Secretsmanager connection."""
        warnings.warn(
            """
            Method `credsman_connect` will be deprecated in future versions.
            Please use `SnowConn.connect(methods=['secretsmanager']` instead
            """,
            DeprecationWarning,
            stacklevel=2,
        )
        return cls.connect_secretsmanager(*args, **kwargs)

    @classmethod
    def connect_credentials(  # noqa: PLR0913
        cls,
        account: str,
        username: str,
        password: str,
        authenticator: str | None = None,
        db: str = "public",
        schema: str = "public",
        autocommit: bool = True,
        role: str | None = None,
        warehouse: str | None = None,
        **kwargs: Any,
    ) -> SnowConn:
        """Creates an engine and connection to snowflake db using the provided credentials."""
        conn = cls()
        creds = {
            "ACCOUNT": account,
            "USERNAME": username,
            "PASSWORD": password,
            "AUTHENTICATOR": authenticator,
        }

        connect_args = kwargs.get("connect_args", {})
        conn._create_engine(creds, db, schema, autocommit, role, warehouse, connect_args)
        return conn

    def _get_secretsmanager_creds(
        self,
        credsman_name: str,
        region_name: str,
        aws_access_key_id: str | None = None,
        aws_secret_access_key: str | None = None,
    ) -> Any:
        import boto3  # noqa

        aws_creds: dict[str, str] = {}
        if aws_access_key_id and aws_secret_access_key:
            aws_creds = {
                "aws_access_key_id": aws_access_key_id,
                "aws_secret_access_key": aws_secret_access_key,
            }

        session = boto3.Session(**aws_creds, region_name=region_name)  # type: ignore
        return load_from_aws_secret(secret_name=credsman_name, session=session)

    def _create_engine(  # noqa: PLR0913
        self,
        creds: dict[str, Any],
        db: str,
        schema: str,
        autocommit: bool = True,
        role: str | None = None,
        warehouse: str | None = None,
        connect_args: dict[str, Any] | None = None,
    ) -> None:
        role = role if role else creds.get("ROLE")
        connect_args = connect_args or {}
        engine_creds = {
            **creds,
            "autocommit": autocommit,
            "schema": schema,
            "warehouse": warehouse,
            "role": role,
            "database": db,
            **connect_args,
        }
        engine = create_snowflake_sa_engine(engine_creds)
        self._alchemy_engine = engine
        self._connection = self._alchemy_engine.connect()
        self._raw_connection = self._connection.connection.connection

    def get_alchemy_engine(self) -> Engine:
        """Returns the sqlalchemy engine.

        Result of a call to either connect() or credsman_connect().
        """
        return self._alchemy_engine  # type: ignore[return-value]

    def get_connection(self) -> Connection:
        """Returns the snowflake SQLAlchemy connection object.

        Result of calling engine.connect() on an SQLAlchemy engine
        """
        return self._connection  # type: ignore[return-value]

    def get_raw_connection(self) -> SnowflakeConnection:
        """Returns the lower level snowflake-connector connection object."""
        return self._raw_connection  # type: ignore[return-value]

    def execute_simple(self, sql: str) -> list[dict[Any, Any]]:
        """Executes a single SQL statement.

        Reads the result set into memory and
        returns an array of dictionaries. This method is for executing single
        queries from which the result set can be fit into memory in regular
        python data structures. This is a wrapper around the snowflake
        connector execute() method found here: https://docs.snowflake.net/manuals/user-guide/python-connector-api.html#execute

        :param sql: string containing a single SQL statement
        :return: array of dictionaries
        """
        types_to_parse = (5, 9, 10)
        try:
            cursor = self.get_raw_connection().cursor(snowflake.connector.DictCursor)
            results = cursor.execute(sql)
        except snowflake.connector.errors.ProgrammingError as e:
            print(sql)
            raise e

        to_parse = {desc[0] for desc in results.description if desc[1] in types_to_parse}  # type: ignore

        return [
            {
                key: json.loads(value) if key in to_parse and value is not None else value
                for key, value in entry.items()  # type: ignore
            }
            for entry in results  # type: ignore
        ]

    def execute_string(self, sql: str, *args: Any, **kwargs: Any) -> Any:
        """Executes a list of sql statements.

        This is a thin wrapper around the snowflake connector execute_string() method found here:
        https://docs.snowflake.net/manuals/user-guide/python-connector-api.html#execute_string

        :param sql:
        :return: list of cursors
        """
        try:
            cursor_list = self.get_raw_connection().execute_string(sql, *args, **kwargs)
        except snowflake.connector.errors.ProgrammingError as e:
            print(sql)
            raise e
        return cursor_list

    def execute_file(self, fname: str) -> Any:
        """Given the path to filename, execute the contents of the file using self.execute_string.

        :param fname: file path that can be open()ed
        :return: list of cursors
        """
        return self.execute_string(Path(fname).read_text())

    def read_df(self, sql: str, lowercase_columns: bool = True) -> pd.DataFrame:
        """Executes the sql passed in and reads the result into a pandas dataframe.

        If you want to use pandas, you'll have to install it yourself as it is
        not a requirement of this package due to its weight.
        :param sql: string containing a single SQL statement
        :param lowercase_columns: boolean, wether or not to lowercase column names
        (snowflake connector fetch_pandas_all returns uppercase)
        :return: pandas DataFrame
        """
        try:
            import pandas as pd  # noqa
        except ImportError as e:
            logging.warning("pandas not installed, cannot execute read_df")
            raise e
        cursor = self.get_raw_connection().cursor(snowflake.connector.DictCursor)
        cursor.execute(sql)
        read_df = cursor.fetch_pandas_all()
        if lowercase_columns:
            read_df.columns = map(str.lower, read_df.columns)  # type: ignore
        return read_df

    def write_df(  # noqa: PLR0913
        self,
        df: pd.DataFrame,
        table: str,
        schema: str | None = None,
        if_exists: Literal["fail", "replace", "append"] = "replace",
        index: bool = False,
        temporary_table: bool = False,
        chunksize: int = 5000,
        **kwargs: Any,
    ) -> None:
        """Writes a dataframe to the specified table.

        Note that you must be connected in the correct context for this to be able to work as you
        cannot specify the fully namespaced version of the table.

        There reason that we have so many params explicit in this function
        is because we want to change their defaults to something more sensible
        for daltix.

        :param df: pandas dataframe to write to the table
        :param table: the name of ONLY the table
        :param schema: The schema of the table
        :param if_exists: forwarded on to DataFrame.to_sql
        :param index: forwarded on to DataFrame.to_sql
        :param temporary_table: Runs a bit of a hack to create temp tables
        :param chunksize: forwarded on to DataFrame.to_sql
        :param kwargs: forwarded on to DataFrame.to_sql
        :return: None
        """
        import pandas as pd  # noqa: PLC0415

        if schema:
            schema = schema.upper()
        table = table.upper()

        schema_table = (('"' + schema + '".') if schema else "") + ('"' + table + '"')

        if not temporary_table:
            df.to_sql(
                table,
                con=self.get_connection(),
                schema=schema,
                if_exists=if_exists,
                index=index,
                chunksize=chunksize,
                **kwargs,
            )
        else:
            sql = pd.io.sql.get_schema(df, name=table, con=self.get_connection()).replace(  # type: ignore
                f'CREATE TABLE "{table}"', f"CREATE OR REPLACE TEMPORARY TABLE {schema_table}"
            )
            self.execute_simple(sql)
            df.to_sql(
                table,
                con=self.get_connection(),
                schema=schema,
                if_exists="append",
                index=index,
                chunksize=chunksize,
                **kwargs,
            )

    def close(self) -> None:
        """Close off the current connection.

        Also, dispose() of the engine is not documented anywhere in snowflake.

        :return: None
        """
        self.get_connection().close()
        self.get_alchemy_engine().dispose()

    def get_current_role(self) -> Any:
        """Get the current role of the connection."""
        results = self.execute_simple("show roles;")
        return next(r for r in results if r["is_current"] == "Y")["name"]



"""Python utilities for connection to the Snowflake data warehouse."""

from __future__ import annotations

from .connect import SnowConn
from .connection_builder import (
    ConnectionError,
    load_from_aws_secret,
    load_from_env_vars,
    load_from_json_file,
    load_from_snowflake_config_file,
)


__all__ = [
    "ConnectionError",
    "SnowConn",
    "load_from_aws_secret",
    "load_from_env_vars",
    "load_from_json_file",
    "load_from_snowflake_config_file",
]
